


class Signal:
    "Google signals and slots to learn the concept."
    def __init__(self):
        self.slots_ = []
    
    def connect(self, func):
        self.slots_.append(func)
        return Connection(self, func)
    
    def disconnect(self, func):
        self.slots_.remove(func)
        
    def emit(self):
        for func in self.slots_:
            func()
            
class Connection:
    "The purpose of this class is to avoid the need to keep track of the origin of func"
    def __init__(self, signal, func):
        self.signal_ = signal
        self.func_ = func
        
    def disconnect(self):
        self.signal_.disconnect(self.func_)

class FrameMover:
    def __init__(self):
        self.pause_ = False
        self.total_elapsed_time_ = 0
        self.signal_on_frame_move_ = Signal()
        self.signal_on_pause_ = Signal()
        self.signal_on_resume_ = Signal()
        
    def onFrameMove(self, t):
        pass
    
    def onPause(self):
        pass
    
    def onResume(self):
        pass
    
    def connect_signal_on_frame_move(self, slot):
        return self.signal_on_frame_move_.connect(slot)
    
    def connect_signal_on_pause(self, slot):
        return self.signal_on_pause_.connect(slot)
    
    def connect_signal_on_resume(self, slot):
        return self.signal_on_resume_.connect(slot)
        
    def frame_move(self, t):
        self.total_elapsed_time_ += t
        self.signal_on_frame_move_.emit()
        self.onFrameMove(t)
        
    def pause(self):
        if not self.pause_:
            self.pause_ = True
            self.signal_on_pause_.emit()
            self.onPause()
    
    def resume(self):
        if self.pause_:
            self.pause_ = False
            self.signal_on_resume_.emit()
            self.onResume()
            
    def toggle_pause(self):
        if self.pause_:
            self.resume()
        else:
            self.pause()
    
    def is_paused(self):
        return self.pause_
    
    def total_elapsed_time(self):
        return self.total_elapsed_time_
    
    def reset_time(self):
        self.total_elapsed_time_ = 0



from .State import State
from .FrameMover import Signal
import gc
import weakref


class TimedEventType:
    def __init__(self, time, strEvent, cancelable):
        self.time_ = time
        self.event_ = strEvent
        self.cancelable_ = cancelable


class StateMachine(State):
    
    def __init__(self, manager):
        self.manager_ = weakref.ref(manager)
        self.slots_prepared_ = False
        self.slots_connected_ = False
        self.scxml_loaded_ = False
        self.engine_started_ = False
        self.on_event_ = False
        self.with_history_ = False
        self.current_active_state_ = None
        self.current_enter_state_ = None
        self.leaf_states_ = []
        self.frame_move_slots_ = dict()
        self.cond_slots_ = dict()
        self.action_slots_ = dict()
        self.allow_nop_entry_exit_slot_ = False
        self.do_exit_state_on_destroy_ = False
        self.transition_source_state_ = self.transition_target_state_ = ''
        self.timed_events_ = []
        self.queued_events_ = []
        self.states_map_ = dict()
        self.signal_prepare_slots_ = Signal();
        self.signal_connect_cond_slots_ = Signal();
        self.signal_connect_action_slots_ = Signal();
        
        State.__init__(self, '', None, self)

        self.addState(self)
    
    
    def __del__(self):
        self.clearTimedEvents()
        self.destroy_machine(self.do_exit_state_on_destroy_)
        State.__del__(self)
        
    def set_do_exit_state_on_destroy(self, yes_no):
        self.do_exit_state_on_destroy_ = yes_no
        
    def clone(self):
        mach = StateMachine(self.manager_())
        mach.state_id_ = self.state_id_
        mach.scxml_id_ = self.scxml_id_
        mach.scxml_loaded_ = self.scxml_loaded_
        
        mach.machine_ = weakref.ref(mach)
        mach.clone_data(self)
        
        return mach
        
        
    def is_unique_id(self, state_id):
        if not self.manager_():
            print('no manager')
            return True
        else:
            return self.manager_().is_unique_id(self.scxml_id_, state_id)
    
    
    def getState(self, state_uid):
        if state_uid in self.states_map_:
            return self.states_map_[state_uid]
        else:
            return None

    
    def addState(self, state):
        if not state:
            return
        self.states_map_[state.state_uid()] = state
        
        
    def removeState(self, state):
        if not state or state is self:
            return
        if self.states_map_[state.state_uid()] is state:
            del self.states_map_[state.state_uid()]
        
        
    def inState(self, state_uid):
        if state_uid in self.states_map_:
            return self.states_map_[state_uid].active()
        else:
            return False

    def elapsed_time_of_last_enter_state(self):
        "elapsed time of the last entry state"
        return self.current_enter_state_.total_elapsed_time_

    def elapsed_time_of_current_state(self):
        "elapsed time of the active state, for use in frame_move."
        return self.current_active_state_.total_elapsed_time_

    def elapsed_time_of_leaf_states(self):
        "if you use parallel, there will be more than 1 leaf states."
        return [state.total_elapsed_time_ for state in self.leaf_states_]

    def onFrameMove(self, t):
        if not self.slots_connected_:
            return
        self.pumpTimedEvents()
        while len(self.queued_events_):
            self.pumpQueuedEvents()
        State.onFrameMove(self, t)
        while len(self.queued_events_):
            self.pumpQueuedEvents()

    def pumpQueuedEvents(self):
        queued_events = self.queued_events_
        self.queued_events_ = []
        for event in queued_events:
            self.onEvent(event)
        
    def enqueEvent(self, e):
        self.queued_events_.append(e)
        self.manager_().addToActiveMach(self)
        
        
    def onEvent(self, e):
        if not self.slots_connected_:
            raise Exception('Slots not connected yet!')
        if self.on_event_:
            self.enqueEvent(e)
            return
        #print(f'onEvent({e})')
        self.on_event_ = True
        State.onEvent(self, e)
        self.on_event_ = False


    def prepareEngine(self):
        if not self.scxml_loaded_:
            raise Exception('no scxml loaded')
        self.prepare_slots()
        self.connect_slots()
        
        
    def StartEngine(self):
        if not self.scxml_loaded_:
            raise Exception('no scxml loaded')
        if self.engine_started_:
            return
        self.prepareEngine()
        self.engine_started_ = True
        self.enterState()
        
        
    def ReStartEngine(self):
        if not self.scxml_loaded_:
            raise Exception('no scxml loaded')
        if self.engine_started_:
            self.ShutDownEngine()
        self.StartEngine()


    def engineStarted(self):
        return engine_started_
    
    
    def ShutDownEngine(self, do_exit_state):
        if do_exit_state:
            self.exitState()
        self.engine_started_ = False
        
        
    def engineReady(self):
        return self.scxml_loaded_;
    
    
    def isLeavingState(self):
        return self.getEnterState().isLeavingState()


    def re_enter_state(self):
        return self.transition_source_state_ == self.transition_target_state_
    

    def getEnterState(self):
        return self.current_enter_state_
    
    
    def prepare_slots(self):
        if self.slots_prepared_:
            return
        
        self.slots_prepared_ = True
        self.prepareActionCondSlots()
        self.onPrepareActionCondSlots()
        self.signal_prepare_slots_.emit()


    def connect_slots(self):
        if self.slots_connected_:
            return
        
        self.slots_connected_ = True
        self.connectCondSlots()
        self.onConnectCondSlots()
        self.signal_connect_cond_slots_.emit()

        self.connectActionSlots()
        self.onConnectActionSlots()
        self.signal_connect_action_slots_.emit()


    def clear_slots(self):
        self.slots_connected_ = False
        self.action_slots_ = dict()
        self.cond_slots_ = dict()
        self.frame_move_slots_ = dict()
        self.transitions_ = []
            

    def destroy_machine(self, do_exit_state):
        if self.slots_connected_:
            if do_exit_state:
                self.exitState()
            self.scxml_loaded_ = False
        self.queued_events_ = []
        self.states_map_ = dict()
        self.machine_clear_substates()
        self.clear_slots()
        self.reset_history()
        
        self.engine_started_ = False


    def loadSCXMLString(self, xmlstr):
        if self.scxml_loaded_:
            self.destroy_machine()
        self.scxml_loaded_ = self.manager_().loadMachFromString(self, xmlstr)
        if not self.scxml_loaded_:
            self.destroy_machine()
        self.onLoadScxmlFailed()
        
        
    def GetCondSlot(self, name):
        if not self.cond_slots_:
            return None
        if name in self.cond_slots_:
            return self.cond_slots_[name]
        return None


    def GetActionSlot(self, name):
        if not self.action_slots_:
            return None
        if name in self.action_slots_:
            return self.action_slots_[name]
        return None


    def GetFrameMoveSlot(self, name):
        if not self.frame_move_slots_:
            return None
        if name in self.frame_move_slots_:
            return self.frame_move_slots_[name]
        return None
    
    
    def setCondSlot(self, name, s):
        self.cond_slots_[name] = s
        
    
    def setActionSlot(self, name, s):
        self.action_slots_[name] = s
        
        
    def setFrameMoveSlot(self, name, s):
        self.frame_move_slots_[name] = s
    

    def registerTimedEvent(self, after_t, event_e, cancelable):
        p = TimedEventType(after_t + self.total_elapsed_time_, event_e, cancelable)
        idx = 0
        for element in self.timed_events_:
            if p.time_ > element.time_:
                idx = i
                break
            idx += 1
        self.timed_events_.insert(idx, p)
        return p
        
        
    def clearTimedEvents(self):
        self.timed_events_.clear()
        
        
    def pumpTimedEvents(self):
        idx = 0
        for p in self.timed_events_:
            if p.time_ <= self.total_elapsed_time_:
                if not p.cancelable_ or len(gc.get_referrers(p)) > 1:
                    self.machine_.enqueEvent(p.event_)
                del self.timed_events_[idx]
            else:
                break
            idx += 1


    def state_id_of_history(self, history_id):
        return self.manager_().history_id_resided_state(self.scxml_id_, history_id)
    
    
    def history_type(self, state_uid):
        return self.manager_().history_type(self.scxml_id_, state_uid)
    
    
    def initial_state_of_state(self, state_uid):
        return self.manager_().initial_state_of_state(self.scxml_id_, state_uid)
    
    
    def onentry_action(self, state_uid):
        return self.manager_().onentry_action(self.scxml_id_, state_uid)
    
    
    def onexit_action(self, state_uid):
        return self.manager_().onexit_action(self.scxml_id_, state_uid)
    
    
    def frame_move_action(self, state_uid):
        return self.manager_().frame_move_action(self.scxml_id_, state_uid)
    
    
    def transition_attr(self, state_uid):
        return self.manager_().transition_attr(self.scxml_id_, state_uid)
    
    
    def num_of_states(self):
        return len(self.states_map_)
        

    def get_all_states(self):
        return self.manager_().get_all_states(self.scxml_id_)

    def manager(self):
        return self.manager_()

    def onPrepareActionCondSlots(self):
        pass
    
    def onConnectCondSlots(self):
        pass
    
    def onConnectActionSlots(self):
        pass
    
    def onLoadScxmlFailed(self):
        pass
    
    def register_state_slot(self, state, onentry, onexit):
        self.setActionSlot('onentry_'+state, onentry)
        self.setActionSlot('onexit_'+state, onexit)

    def register_action_slot(self, action, method):
        self.setActionSlot(action, method)
        
    def register_frame_move_slot(self, state, method):
        self.setFrameMoveSlot(state, method)
        
    def register_cond_slot(self, cond, method):
        self.setCondSlot(cond, method)

    def register_handler(self, handler):
        states = self.get_all_states()
        for state in states:
            s = state.replace('.', '_')
            try:
                self.setActionSlot('onentry_'+state, eval('handler.onentry_' + s))
            except:
                pass
            try:
                self.setActionSlot('onexit_'+state, eval('handler.onexit_' + s))
            except:
                pass
        
        
        
        
        



from .FrameMover import FrameMover, Signal, Connection
from random import randint
import weakref

class TransitionAttr:
    def __init__(self, event, target):
        self.event_ = event
        self.transition_target_ = target
        self.not_ = False   # to support "!In(state)"
        self.random_target_ = []
        self.cond_ = ''     # for later connecting slot
        self.ontransit_ = ''# for later connecting slot
        self.in_state_ = [] # for inState check if set. You can use '|' to specify multiple states, ex. "In(state1|state2)"
        

class Transition:
    def __init__(self, attr):
        self.attr_ = attr
        self.cond_functor_ = None
        self.signal_transit_ = Signal()
        
    def setAttr(self, attr):
        self.attr_ = attr
        
        
class State(FrameMover):

    done_state_prefix = "done.state."
    

    def __init__(self, state_id, parent, machine):
        FrameMover.__init__(self)
        self.parent_ = weakref.ref(parent) if parent else None
        self.machine_ = weakref.ref(machine) if machine else None
        self.substates_  = []
        self.no_event_transitions_ = []
        self.transitions_ = []
        self.frame_move_slots_ = []
        self.slots_ready_ = False
        if self.parent_:
            self.depth_ = self.parent_().depth_ + 1
        else:
            self.depth_ = 0
        self.active_ = False
        self.done_ = False
        self.is_a_final_ = False
        self.current_state_ = None
        self.history_state_id_ = ''
        self.leaving_delay_ = 0.0
        self.leaving_target_transition_ = None
        self.leaving_elapsed_seconds_ = 0.0
        self.state_id_ = ''
        self.is_unique_state_id_ = True
        self.set_state_id(state_id)
        self.signal_done_ = Signal()
        self.signal_onentry_ = Signal()
        self.signal_onexit_ = Signal()
        #print(f'new {self.state_uid_}')
    
    def __del__(self):
        #print(f'delete {self.state_uid_}')
        pass

    def set_state_id(self, state_id):
        if self.slots_ready_:
            raise Exception("slots already ready, can't change state id!")

        if self.state_id_ and self.machine_() is not self:
            self.machine_().removeState(self)
            
        self.state_id_ = state_id
        
        if not self.state_id_: # anonymous state, generate one
            if self.machine_() is self:
                self.state_id_ = '_root'
            else:
                self.state_id_ = '_st' + str(self.machine_().num_of_states())
        else:
            self.is_unique_state_id_ = self.machine_().is_unique_id(self.state_id_)
            
        if self.is_unique_state_id_:
            self.state_uid_ = self.state_id_
        else:
            self.state_uid_ = self.parent_().state_uid_ + '.' + self.state_id_
            
        if self.machine_() is not self: # delay addState to after machine_() construction complete.
            self.machine_().addState(self)
            

    def state_id(self):
        return self.state_id_
    

    def state_uid(self):
        return self.state_uid_
    

    def depth(self):
        return self.depth_
    

    def active(self):
        return self.active_
    

    def done(self):
        return self.done_
    
    
    def leavingDelay(self):
        return self.leaving_delay_


    def isLeavingState(self):
        return self.leaving_target_transition_
    
    
    def setLeavingDelay(self, delay):
        self.leaving_delay_ = delay
    

    def initial_state(self):
        initstate = self.machine_().initial_state_of_state(self.state_uid_)
        if initstate:
            return inits;
        elif self.substates_:
            return substates_[0].state_uid()
        else:
            return "" # a leaf state has no initial.
        

    def machine_clear_substates(self):
        for substate in self.substates_:
            substate.machine_clear_substates ()

        self.machine_ = None
        self.substates_ = []


    def clone(self, parent, machine):
        state = State(self.state_id_, parent, machine)
        state.clone_data(self)
        return state

    
    def clone_data(self, rhs):
        for substate in rhs.substates_:
            self.substates_.append(substate.clone(self, self.machine_()))
            
        self.depth_ = rhs.depth_
        self.is_a_final_ = rhs.is_a_final_
        self.is_unique_state_id_ = rhs.is_unique_state_id_
        self.leaving_delay_ = rhs.leaving_delay_

        
    def reset_history(self):
        if not self.machine_:
            return
        
        if self.machine_().history_type(self.state_uid_):
            self.clearHistory()
            
        for substate in self.substates_:
            substate.reset_history()


    def clearHistory(self):
        self.history_state_id_ = ''

        
    def clearDeepHistory(self):
        self.history_state_id_ = ''
        for substate in self.substates_:
            substate.clearDeepHistory()

            
    def trig_cond(self, tran):
        if tran.cond_functor_:
            change = tran.cond_functor_()
            if tran.attr_.not_:
                change = not change
            return change
        elif tran.attr_.in_state_:
            change = False
            for s in tran.attr_.in_state_:
                change |= self.machine_().inState(s)
            if tran.attr_.not_:
                change = not change
            return change
        else:
            return True


    def enterState(self, enter_substate=True):
        if self.active_:
            return # already in state
        
        if self.parent_ and self.machine_().history_type(self.parent_().state_uid()):
            self.parent_().history_state_id_ = self.state_uid_
            
        self.reset_time()

        self.done_ = False
        self.active_ = True
        
        self.machine_().current_enter_state_ = self

        if not self.substates_:
            self.machine_().leaf_states_.append(self)

        self.signal_onentry_.emit()
        
        if not self.active_: # in case state changed immediately at last signal_onentry.
            return
        
        if enter_substate:
            self.doEnterSubState()
            
        if self.is_a_final_ and self.parent_():
            self.parent_().done_ = True
            self.parent_().signal_done_.emit()
            self.machine_().enqueEvent(State.done_state_prefix + self.parent_().state_uid())


    def exitState(self):
        if not self.active_:
            return
        #print(f'exitState {self.state_uid()}, mach {self.machine_}')
        if len(self.substates_) == 0 and self.machine_ and self.machine_():
            self.machine_().leaf_states_.remove(self)
        elif self.current_state_:
            self.current_state_.exitState()
            
        self.active_ = False
        self.current_state_ = None
        
        self.signal_onexit_.emit()

    
    def onEvent(self, e):
        if self.done_:
            return
        
        if self.isLeavingState():
            return
        
        for tran in self.transitions_:
            if tran.attr_.event_ == e:
                change = self.trig_cond(tran)
                if change:
                    self.changeState(tran)
                    return
                
        if self.current_state_:
            self.current_state_.onEvent(e)


    def findState(self, state_id, exclude=None, check_parent=True):
        for substate in self.substates_:
            if substate.state_uid() == state_id or substate.state_id() == state_id:
                return substate
            
        for substate in self.substates_:
            if substate is not exclude:
                st = substate.findState(state_id, exclude, False)
                if st:
                    return st
                
        if check_parent and self.parent_():
            st = self.parent_().findState(state_id, self, True)
            if st:
                return st
            
        return None


    def changeState(self, transition):
        if transition.attr_.random_target_:
            index = randint(0, len(transition.attr_.random_target_)-1)
            target = transition.attr_.random_target_[index]
        else:
            target = transition.attr_.transition_target_
            
        corresponding_state = self.machine_().state_id_of_history(target)
        if corresponding_state:
            st = self.machine_().getState(corresponding_state)
            if st.history_state_id_:
                target = st.history_state_id_
            else:
                target = st.initial_state()
                
        if not self.leaving_target_transition_:
            if self.leaving_delay_ != 0:
                attr = TransitionAttr(transition.attr_.event_, target)
                self.leaving_target_transition_ = Transition(attr)
                self.leaving_target_transition_.setAttr(attr)
                if transition.attr_.ontransit_:
                    slot = self.machine_().GetActionSlot(transition.attr_.ontransit_)
                    if slot:
                        self.leaving_target_transition_.signal_transit_.connect(slot)
                return
                
        newState = []
        if ',' not in target:
            st = self.machine_().getState(target)
            if st:
                newState.append(st)
            else:
                raise Exception(f"can't find transition target '{target}'!")
        else:
            targets = target.split(',')
            for t in targets:
                st = self.machine_().getState(t)
                if st:
                    newState.append(st)
                else:
                    raise Exception("can't find transition target state!")

        if not newState:
            raise Exception("can't find transition target state!")

        lcaStates = []
        lcaState = None
        for state in newState:
            max_depth = 0
            lca = None
            for leaf_state in self.machine_().leaf_states_:
                lca = leaf_state.findLCA(state)
                if lca.depth_ >= max_depth:
                    max_depth = lca.depth_
                    lcaState = lca
            if lcaState is None:
                raise Exception("can't find common ancestor state.")
            lcaStates.append(lcaState)

        if len(newState) == 1 and lcaState is newState[0]: # reentering state
            self.machine_().transition_source_state_ = lcaState.state_uid()
            self.machine_().transition_target_state_ = lcaState.state_uid()
            lcaState.exitState()
            transition.signal_transit_.emit()
            lcaState.enterState()
            return

        for uState, lcaState in zip(newState, lcaStates):
            self.machine_().transition_target_state_ = uState.state_uid()
            if lcaState.current_state_ is not None:
                self.machine_().transition_source_state_ = lcaState.state_uid()
                lcaState.current_state_.exitState()
            else: # is a parallel state
                for substate in lcaState.substates_:
                    if substate.isAncestorOf(uState):
                        self.machine_().transition_source_state_ = substate.state_uid()
                        substate.exitState()
                        break
            transition.signal_transit_.emit()

            # enter new state
            vps = [uState]
            parentstate = uState.parent_()
            while parentstate and parentstate is not lcaState:
                vps.append(parentstate)
                parentstate = parentstate.parent_()

            lcaState.doEnterState(vps)


    def doEnterState(self, vps):
        state = vps[-1]
        if state.depth_ <= self.depth_:
            return
        
        vps.pop()
        self.current_state_ = state
        enter_sub = len(vps) == 0 or vps[-1].depth_ <= self.current_state_.depth_;
        self.current_state_.enterState(enter_sub)
        if vps:
            self.current_state_.doEnterState(vps)
        

    def doEnterSubState(self):
        if self.history_state_id_ and self.machine_().history_type(self.state_uid_):
            self.current_state_ = self.machine_().getState(self.history_state_id_)
            self.current_state_.enterState()
        else:
            if self.substates_:
                initial_state = self.machine_().initial_state_of_state(self.state_uid_)
                if not initial_state:
                    self.current_state_ = self.substates_[0]
                    self.current_state_.enterState()
                else:
                    attr = TransitionAttr('', initial_state)
                    tran = Transition(attr)
                    self.changeState(tran)


    def findLCA(self, ots):
        if not ots:
            raise Exception("findLCA on None object")
        if self is ots:
            return self
        elif self.depth_ > ots.depth_:
            return self.parent_().findLCA(ots)
        elif self.depth_ < ots.depth_:
            return self.findLCA(ots.parent_())
        else:
            if self.parent_() is ots.parent_():
                return self.parent_()
            else:
                return self.parent_().findLCA(ots.parent_())
        
        return None

    def isAncestorOf(self, ots):
        while ots is not None and ots.depth_ >= self.depth_:
            if ots is self:
                return True
            ots = ots.parent_
        return False

    def inState(self, state_id, recursive=True):
        if not self.current_state_:
            return False
        if self.current_state_.state_id_ == state_id:
            return True
        elif recursive:
            return self.current_state_.inState(state_id, recursive)
        else:
            return False


    def inStateObj(self, state, recursive=True):
        if not self.current_state_:
            return False
        if self.current_state_ is state:
            return True
        elif recursive:
            return self.current_state_.inStateObj(state, recursive)
        else:
            return False


    def prepareActionCondSlots(self):
        if self.slots_ready_:
            return
        
        self.slots_ready_ = True
        
        tran_attrs = self.machine_().transition_attr(self.state_uid_)
        size = len(tran_attrs)
        for i in range(size):
            ptr = Transition(tran_attrs[i])
            if not tran_attrs[i].event_:
                self.no_event_transitions_.append(ptr)
            else:
                self.transitions_.append(ptr)
        
        self.machine_().setActionSlot(f'clh({self.state_uid_}*)', self.clearDeepHistory)
        self.machine_().setActionSlot(f'clh({self.state_uid_})', self.clearHistory)
        
        for substate in self.substates_:
            substate.prepareActionCondSlots()


    def connectCondSlots(self):
        self.connect_transitions_conds(self.transitions_)
        self.connect_transitions_conds(self.no_event_transitions_)
        
        for substate in self.substates_:
            substate.connectCondSlots()


    def connect_transitions_conds(self, transitions):
        """      """
        size = len(transitions)
        for i in range(size):
            if transitions[i].attr_.cond_:
                cond = transitions[i].attr_.cond_
                instate_check = cond[0:3]
                if instate_check == "In(" or instate_check == "in(":
                    endmark = cond.find(')', 4)
                    states = cond[3:endmark].split("|")
                    for st in states:
                        if not self.machine_().is_unique_id(st):
                            s = self.findState(st)
                            if not s:
                                raise Exception("can't find state for In() check")
                            st = s.state_uid()
                        else:
                            s = self.findState(st)
                            if not s:
                                raise Exception("can't find state for In() check")
                        transitions[i].attr_.in_state_.append(st)
                else:
                    s = self.machine_().GetCondSlot(cond)
                    if s:
                        transitions[i].cond_functor_ = s
                    else:
                        raise Exception("can't connect cond slot")
        

    def connectActionSlots(self):
        onentry = self.machine_().onentry_action(self.state_uid_)
        if onentry:
            s = self.machine_().GetActionSlot(onentry)
            if s:
                self.signal_onentry_.connect(s)
            elif onentry != f'onentry_{self.state_uid_}':
                raise Exception("can't connect onentry slot")

        onexit = self.machine_().onexit_action(self.state_uid_)
        if onexit:
            s = self.machine_().GetActionSlot(onexit)
            if s:
                self.signal_onexit_.connect(s)
            elif onexit != f'onexit_{self.state_uid_}':
                raise Exception("Can't connect onexit slot")
            
        frame_move = self.machine_().frame_move_action(self.state_uid_)
        if frame_move:
            sf = self.machine_().GetFrameMoveSlot(frame_move)
            if sf:
                self.frame_move_slots_.append(sf)
            elif frame_move != self.state_uid_:
                raise Exception("Can't connect frame_move slot")
            
        self.connect_transitions_signal(self.transitions_)
        self.connect_transitions_signal(self.no_event_transitions_)
        
        for substate in self.substates_:
            substate.connectActionSlots()


    def connect_transitions_signal(self, transitions):
        for tran in transitions:
            if tran.attr_.ontransit_:
                tr = tran.attr_.ontransit_
                s = self.machine_().GetActionSlot(tr)
                if s:
                    tran.signal_transit_.connect(s)
                else:
                    raise Exception("can't connect on_transit slot")
                

    def doLeaveAferDelay(self):
        if self.leaving_target_transition_:
            self.changeState(self.leaving_target_transition_)
            self.leaving_target_transition_ = None
            self.leaving_elapsed_seconds_ = 0

    def check_leaving_state_finished(self, t):
        if self.leaving_delay_ < 0:
            return False
        self.leaving_elapsed_seconds_ += t
        if self.leaving_elapsed_seconds_ >= self.leaving_delay_:
            self.doLeaveAferDelay()
            return True
        return False
        

    def onFrameMove(self, t):
        if self.isLeavingState():
            if self.check_leaving_state_finished(t):
                return
            
        if self.current_state_:
            self.current_state_.frame_move(t)
            if not self.active_:
                return 
    
        self.pumpNoEvents()
        
        if not self.active_:
            return
        
        self.machine_().current_active_state_ = self
        for slot in self.frame_move_slots_:
            slot(t)
            
            
    def pumpNoEvents(self):
        if self.isLeavingState():
            return
        
        for tran in self.no_event_transitions_:
            if self.trig_cond(tran):
                self.changeState(tran)
                return




#!/usr/bin/env python3
import sys

from setuptools import find_packages, setup


# Get version without importing, which avoids dependency issues

def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='py-scm',
      version='1.0.5',
      description='A python state machine framework based on statecharts (scxml)',
      long_description=readme(),
      author='Zen Chien',
      author_email='jixing.jian@gmail.com',
      url='https://github.com/zen747/pyscm',
      license="Apache License 2.0",
      keywords=['statecharts', 'state-machine', 'scxml', 'david-harel'],
      classifiers=["Development Status :: 5 - Production/Stable",
                   "Intended Audience :: Developers",
                   "License :: OSI Approved :: Apache Software License",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.6',
                   "Topic :: Software Development :: Libraries :: Python Modules",
                   "Topic :: Text Processing :: Linguistic"],
      packages=find_packages(),
      python_requires=">=3.6")



from .StateMachine import StateMachine
from .State import State, TransitionAttr
from .Parallel import Parallel
import xml.etree.ElementTree as etree
import json

class ParseStruct:
    def __init__(self):
        self.current_state_ = 0
        self.machine_ = 0
        self.scxml_id_ = ''


class StateMachineManager:
    def __init__(self):
        self.mach_map_ = dict()
        self.scxml_map_ = dict()
        self.onentry_action_map_ = dict()
        self.onexit_action_map_ = dict()
        self.frame_move_action_map_ = dict()
        self.initial_state_map_ = dict()
        self.history_type_map_ = dict()
        self.history_id_reside_state_ = dict()
        self.transition_attr_map_ = dict()
        self.state_uids_ = dict()
        self.non_unique_ids_ = dict()
        self.active_machs_ = []
        
    def getMach(self, scxml_id):
        if scxml_id in self.mach_map_:
            return self.mach_map_[scxml_id].clone()
        else:
            mach = StateMachine(self)
            self.mach_map_[scxml_id] = mach
            mach.scxml_id_ = scxml_id
            if scxml_id in self.scxml_map_:
                if self.scxml_map_[scxml_id].startswith('file:'):
                    mach.scxml_loaded_ = self.loadMachFromFile(mach, self.scxml_map_[scxml_id][5:])
                else:
                    mach.scxml_loaded_ = self.loadMachFromString(mach, self.scxml_map_[scxml_id])
            return mach.clone()

    def addToActiveMach(self, mach):
        if not mach:
            raise Exception('mach None?')
        self.active_machs_.append(mach)
        
    def pumpMachEvents(self):
        while self.active_machs_:
            machs = self.active_machs_
            self.active_machs_ = []
            for mach in machs:
                mach.pumpQueuedEvents()

    def set_scxml(self, scxml_id, scxml_str):
        self.scxml_map_[scxml_id] = scxml_str
        
    def set_scxml_file(self, scxml_id, scxml_filepath):
        self.scxml_map_[scxml_id] = 'file:' + scxml_filepath
        
    def prepare_machs(self):
        for scxml_id in self.scxml_map_:
            mach = None
            if scxml_id in self.mach_map_:
                mach = self.mach_map_[scxml_id]
            else:
                mach = StateMachine(self)
                mach.scxml_id_ = scxml_id
                self.mach_map_[scxml_id] = mach
            if not mach.scxml_loaded_:
                if self.scxml_map_[scxml_id].startswith('file:'):
                    mach.scxml_loaded_ = self.loadMachFromFile(mach, self.scxml_map_[scxml_id][5:])
                else:
                    mach.scxml_loaded_ = self.loadMachFromString(mach, self.scxml_map_[scxml_id])
                
    def loadMachFromFile(self, mach, scxml_file):
        with open(scxml_file, 'r') as f:
            self.loadMachFromString(mach, f.read())
            
    def loadMachFromString(self, mach, scxml_str):
        parset_data = ParseStruct()
        parset_data.scxml_id_ = scxml_id = mach.scxml_id_
        parset_data.machine_ = mach
        parset_data.current_state_ = mach
        self.transition_attr_map_[scxml_id] = dict()
        self.state_uids_[scxml_id] = []
        self.non_unique_ids_[scxml_id] = []
        self.onentry_action_map_[scxml_id] = dict()
        self.onexit_action_map_[scxml_id] = dict()
        self.frame_move_action_map_[scxml_id] = dict()
        self.initial_state_map_[scxml_id] = dict()
        self.history_type_map_[scxml_id] = dict()
        self.history_id_reside_state_[scxml_id] = dict()
        
        return self.parse_scm_tree(parset_data, scxml_str)
        
    def clearMachMap(self):
        self.mach_map_.clear()
        self.onentry_action_map_.clear()
        self.onexit_action_map_.clear()
        self.frame_move_action_map_.clear()
        self.initial_state_map_.clear()
        self.history_type_map_.clear()
        self.history_id_reside_state_.clear()
        self.transition_attr_map_.clear()
        
    def history_id_resided_state(self, scxml_id, history_id):
        try:
            return self.history_id_reside_state_[scxml_id][history_id]
        except KeyError:
            return ''
    
    def history_type(self, scxml_id, state_uid):
        try:
            return self.history_type_map_[scxml_id][state_uid]
        except KeyError:
            return ''
        
    def initial_state_of_state(self, scxml_id, state_uid):
        try:
            return self.initial_state_map_[scxml_id][state_uid]
        except KeyError:
            return ''
        
    def onentry_action(self, scxml_id, state_uid):
        try:
            return self.onentry_action_map_[scxml_id][state_uid]
        except KeyError:
            return ''

    def onexit_action(self, scxml_id, state_uid):
        try:
            return self.onexit_action_map_[scxml_id][state_uid]
        except KeyError:
            return ''
    
    def frame_move_action(self, scxml_id, state_uid):
        try:
            return self.frame_move_action_map_[scxml_id][state_uid]
        except KeyError:
            return ''
    
    def transition_attr(self, scxml_id, state_uid):
        try:
            return self.transition_attr_map_[scxml_id][state_uid]
        except KeyError:
            return []
    
    def num_of_states(self, scxml_id):
        return len(self.state_uids_[scxml_id])
    
    def is_unique_id(self, scxml_id, state_uid):
        return state_uid not in self.non_unique_ids_[scxml_id]
    
    def get_all_states(self, scxml_id):
        return self.state_uids_[scxml_id]
        
    def parse_scm_tree (self, data, scm_str):
        scm_str = scm_str.strip()
        try:
            if scm_str[0] == '<':
                self.parse_from_xml(scm_str, data)
            else:
                scm_str = scm_str.replace("'", '"')
                self.parse_from_json(scm_str, data)
        except:
            print('parse scxml failed!')
            raise
            #return False
        
        return True
        
    def parse_from_json(self, scm_str, data):
        'todo: add json support'
        pass
        
    def parse_from_xml(self, scm_str, data):
        root = etree.fromstring(scm_str)
        self.parse_element_xml(data, root, 0)
            
    def validate_state_id(self, stateid):
        if stateid and stateid.startswith('_'):
            raise Exception("state id can't start with '_', which is reserved for internal use.")
        
    def parse_element_xml(self, data, element, level):
        manager = data.machine_.manager()
        scxml_id = data.scxml_id_

        if element.tag == 'scxml':
            manager.state_uids_[scxml_id].append(scxml_id)
            if 'non-unique' in element.attrib:
                non_unique_ids = element.attrib['non-unique'].split(',')
                manager.non_unique_ids_[scxml_id] += non_unique_ids
            
        elif element.tag == 'state':
            stateid = element.attrib['id'] if 'id' in element.attrib else ''
            self.validate_state_id(stateid)
            state = State(stateid, data.current_state_, data.machine_)
            data.current_state_.substates_.append(state)
            data.current_state_ = state
            manager.state_uids_[scxml_id].append(state.state_uid())
            self.handle_state_item(data, element.attrib)
        elif element.tag == 'parallel':
            stateid = element.attrib['id'] if 'id' in element.attrib else ''
            self.validate_state_id(stateid)
            state = Parallel(stateid, data.current_state_, data.machine_)
            data.current_state_.substates_.append(state)
            data.current_state_ = state
            manager.state_uids_[scxml_id].append(state.state_uid())
            self.handle_state_item(data, element.attrib)
        elif element.tag == 'final':
            stateid = element.attrib['id'] if 'id' in element.attrib else ''
            self.validate_state_id(stateid)
            state = State(stateid, data.current_state_, data.machine_)
            data.current_state_.substates_.append(state)
            data.current_state_ = state
            manager.state_uids_[scxml_id].append(state.state_uid())
            self.handle_final_item(data, element.attrib)
        elif element.tag == 'history':
            self.handle_history_item(data, element.attrib)
        elif element.tag == 'transition':
            self.handle_transition_item(data, element.attrib)
            
        for child in element:
            self.parse_element_xml(data, child, level+1)
            
        if element.tag == 'scxml':
            self.finish_scxml(data, element.attrib)
            #data.current_state_ = data.current_state_.parent_()
        elif element.tag == 'state':
            data.current_state_ = data.current_state_.parent_()
        elif element.tag == 'parallel':
            data.current_state_ = data.current_state_.parent_()
        elif element.tag == 'final':
            data.current_state_ = data.current_state_.parent_()
    
    
    @classmethod
    def finish_scxml(self, data, attrib):
        manager = data.machine_.manager()
        scxml_id = data.scxml_id_

        if 'initial' in attrib:
            manager.initial_state_map_[scxml_id][data.current_state_.state_uid()] = attrib['initial']
            
        # process multiple targets transition. 
        # replace non-unique target to unique id
        transition_map = manager.transition_attr_map_[scxml_id]
        for state_uid,transitions in transition_map.items():
            st = data.machine_.getState(state_uid)
            for transition in transitions:
                target_str = transition.transition_target_
                if ',' not in target_str and data.machine_.is_unique_id(target_str):
                    continue
                tstates = []
                targets = target_str.split(',')
                for i in range(len(targets)):
                    if not data.machine_.is_unique_id(targets[i]):
                        s = st.findState(targets[i])
                        assert(s and "can't find transition target, not state id?")
                        targets[i] = s.state_uid()
                    tstates.append(data.machine_.getState(targets[i]))
                transition.transition_target_ = ','.join(targets)
                
                # transition to multiple targets only apply to parallel states.
                for i in range(len(tstates)-1):
                    lca = tstates[0].findLCA(tstates[i+1])
                    assert(isinstance(lca, Parallel) and "multiple targets but can't find common ancestor.")

    @classmethod
    def handle_state_item(self, data, attributes):
        manager = data.machine_.manager()
        scxml_id = data.scxml_id_
        
        onentry = ''
        onexit = ''
        framemove = ''
        history_type = ''
        leaving_delay = 0
        
        for key,value in attributes.items():
            if key == 'id':
                pass
            elif key == 'initial':
                manager.initial_state_map_[scxml_id][data.current_state_.state_uid()] = value
            elif key == 'history':
                history_type = value
            elif key == 'onentry':
                onentry = value
            elif key == 'onexit':
                onexit = value
            elif key == 'frame_move':
                framemove = value
            elif key == 'leaving_delay':
                leaving_delay = float(value)
            else:
                print(f'attribute "{key}" not supported')
                assert (0 and f'attribute "{key}" not supported')
                
        state_uid = data.current_state_.state_uid()
        if leaving_delay: data.current_state_.setLeavingDelay(leaving_delay)
        if not onentry: onentry = "onentry_" + state_uid
        if not onexit: onexit = "onexit_" + state_uid
        if not framemove: framemove = state_uid
        
        if data.current_state_.parent_().state_uid() in manager.history_type_map_[scxml_id] and manager.history_type_map_[scxml_id][data.current_state_.parent_().state_uid()] == 'deep':
            manager.history_type_map_[scxml_id][state_uid] = "deep"
        else:
            manager.history_type_map_[scxml_id][state_uid] = history_type
            
        if not manager.history_type_map_[scxml_id][state_uid]:
            data.machine_.with_history_ = True
            
        manager.onentry_action_map_[scxml_id][state_uid] = onentry
        manager.onexit_action_map_[scxml_id][state_uid] = onexit
        manager.frame_move_action_map_[scxml_id][state_uid] = framemove


    @classmethod
    def handle_final_item(self, data, attributes):
        manager = data.machine_.manager()
        scxml_id = data.scxml_id_
        state_uid = data.current_state_.state_uid()
        onentry = ''
        onexit = ''
        framemove = ''

        for key,value in attributes.items():
            if key == 'id':
                pass
            elif key == 'onentry':
                onentry = value
            elif key == 'frame_move':
                framemove = value
            else:
                assert (0 and f'attribute "key" not supported in "final" state')
        
        if not onentry: onentry = "onentry_" + state_uid
        if not framemove: framemove = state_uid
        
        manager.onentry_action_map_[scxml_id][state_uid] = onentry
        manager.frame_move_action_map_[scxml_id][state_uid] = framemove
        
        data.current_state_.is_a_final_ = True
        
    @classmethod
    def handle_transition_item(self, data, attributes):
        manager = data.machine_.manager()
        scxml_id = data.scxml_id_
        state_uid = data.current_state_.state_uid()

        tran = TransitionAttr('','')

        for key, value in attributes.items():
            if key == 'event':
                tran.event_ = value
            elif key == 'cond':
                tran.cond_ = value
            elif key == 'ontransit':
                tran.ontransit_ = value
            elif key == 'target':
                tran.transition_target_ = value
            elif key == 'random_target':
                tran.random_target_ = value.split(',')

        if tran.cond_ and tran.cond_.startswith('!'):
            tran.cond_ = tran.cond_[1:]
            tran.not_ = True
            
        if state_uid not in manager.transition_attr_map_[scxml_id]:
            manager.transition_attr_map_[scxml_id][state_uid] = []
        manager.transition_attr_map_[scxml_id][state_uid].append(tran)
        
        
    @classmethod
    def handle_history_item(self, data, attributes):
        manager = data.machine_.manager()
        scxml_id = data.scxml_id_
        state_uid = data.current_state_.state_uid()
    
        for key, value in attributes.items():
            if key == 'type':
                manager.history_type_map_[scxml_id][state_uid] = value
            elif key == 'id':
                manager.history_id_reside_state_[scxml_id][value] = state_uid
    
    instance_ = None
    
    @classmethod
    def instance(self):
        if not self.instance_:
            self.instance_ = StateMachineManager()
        return self.instance_

    @classmethod
    def release_instance(self):
        self.instance_ = None












from .State import State


done_state_prefix = "done.state."


class Parallel(State):
    
    def __init__(self, state_id, parent, machine):
        State.__init__(self, state_id, parent, machine)
        self.finished_substates_ = set()
    
    def clone(self, parent, m):
        pa = Parallel(self.state_id_, parent, m)
        pa.clone_data(self)
        return pa
    
    def onEvent(self, e):
        if self.isLeavingState():
            return
        
        if self.done_:
            return
        
        if e[0:len(done_state_prefix)] == done_state_prefix:
            state_uid = e[len(done_state_prefix):]
            if state_uid:
                self.finished_substates_.add(state_uid)
                              
        for tran in self.transitions_:
            if tran.attr_.event_ == e:
                if self.trig_cond(tran):
                    self.changeState(tran)
                    return
                
        for substate in self.substates_:
            substate.onEvent(e)
            
        if len(self.finished_substates_) == len(self.substates_):
            self.done_ = True
            self.signal_done_.emit()
            self.machine_().enqueEvent(done_state_prefix + self.state_uid_)

    
    def enterState(self, enter_substate=True):
        if self.active_:
            return
        
#        if not self.substates_:
#            raise Exception("Parallen have no substates!")
 
        if self.parent_() and self.machine_().history_type(self.parent_().state_uid()):
            self.parent_().history_state_id_ = self.state_uid_
            
        self.finished_substates_ = set()
        
        self.done_ = False
        self.active_ = True
        self.reset_time()
        
        self.machine_().current_enter_state_ = self

        if not self.substates_:
            self.machine_().leaf_states_.append(self)

        self.signal_onentry_.emit()
        
        if self.parent_() and not self.parent_().inStateObj(self, False):
            return
        
        if enter_substate:
            for substate in self.substates_:
                substate.enterState(enter_substate)
        

    def doEnterState(self, vps):
        state = vps[-1]
        if state.depth_ <= self.depth_:
            return
        
        vps.pop()
        
        for substate in self.substates_:
            if state is substate:
                state.enterState(False)
                if vps:
                    state.doEnterState(vps)
                return
        

    def exitState(self):
        if not self.active_:
            return
        
        if len(self.substates_) == 0 and self.machine_ and self.machine_():
            self.machine_().leaf_states_.remove(self)
        else:
            for substate in self.substates_:
                substate.exitState()
            
        self.active_ = False
        
        self.signal_onexit_.emit()
        

    def inState(self, state_id, recursive=True):
        for substate in self.substates_:
            if substate.state_uid_ == state_id:
                return True
            
        if recursive:
            for substate in self.substates:
                if substate.inState(state_id, recursive):
                    return True
                
        return False

    
    def inStateObj(self, state, recursive=True):
        for substate in self.substates_:
            if substate is state:
                return True
            
        if recursive:
            for substate in self.substates_:
                if substate.inStateObj(state, recursive):
                    return True
                
        return False
    
    
    def onFrameMove(self, t):
        if self.isLeavingState():
            if self.check_leaving_state_finished(t):
                return
        
        for substate in self.substates_:
            substate.frame_move(t)
            if not self.active_:
                return
            
        self.pumpNoEvents()
        
        if not self.active_:
            return
        
        self.machine_().current_active_state_ = self
        for slot in self.frame_move_slots_:
            slot(t)


